#include <twil_controllers/cart_linearizing_controller.h>
#include <pluginlib/class_list_macros.h>

#include <kdl/frames.hpp>
#include <kdl_parser/kdl_parser.hpp>

namespace twil_controllers
{	
	CartLinearizingController::CartLinearizingController(void):
		wheelRadius(2)
	{
	}
	
	CartLinearizingController::~CartLinearizingController(void)
	{
		sub_command_.shutdown();
	}
		
	bool CartLinearizingController::init(hardware_interface::EffortJointInterface *robot, ros::NodeHandle &n)
	{
		node_=n;
		robot_=robot;
		
		XmlRpc::XmlRpcValue joint_names;
		if(!node_.getParam("joints",joint_names))
		{
			ROS_ERROR("No 'joints'  parameter in controller. (namespace: %s)",node_.getNamespace().c_str());
			return false;
		}
		
		if(joint_names.getType() != XmlRpc::XmlRpcValue::TypeArray)
		{
			ROS_ERROR("The 'joints' parameter is not a struct. (namespace: %s)",node_.getNamespace().c_str());
			return false;
		}
		
		for(int i=0; i < joint_names.size();i++)
		{
			XmlRpc::XmlRpcValue &name_value=joint_names[i];
			if(name_value.getType() != XmlRpc::XmlRpcValue::TypeString)
			{
				ROS_ERROR("Array of joint names should contain only strings. (namespace: %s)",node_.getNamespace().c_str());
				return false;
			}
			
			hardware_interface::JointHandle j=robot->getHandle((std::string)name_value);
			joints_.push_back(j);
		}
		sub_command_ = node_.subscribe("command",1000,&CartLinearizingController::commandCB,this);
		
		std::string robot_desc_string;
		if(!node_.getParam("/robot_description",robot_desc_string))
                {
			ROS_ERROR("Could not find '/robot_description'.");
			return false;
		}
		
		KDL::Tree tree;
		if (!kdl_parser::treeFromString(robot_desc_string,tree))
		{
			ROS_ERROR("Failed to construct KDL tree.");
			return false;
		}
		
		// get wheelBase from URDF (actually from KDL tree)
		KDL::SegmentMap::const_iterator segmentMapIter=tree.getSegment("left_wheel_support");
		KDL::Frame leftSupportFrame=segmentMapIter->second.segment.getFrameToTip();
                
                segmentMapIter=tree.getSegment("right_wheel_support");
                KDL::Frame rightSupportFrame=segmentMapIter->second.segment.getFrameToTip();
               
		wheelRadius.resize(joints_.size());                
                wheelBase=leftSupportFrame(1,3)-rightSupportFrame(1,3);

		// get wheelRadius from URDF (actually from KDL tree)
		segmentMapIter=tree.getSegment("chassis");
                KDL::Frame chassisFrame=segmentMapIter->second.segment.getFrameToTip();
                
                segmentMapIter=tree.getSegment("left_wheel");
                KDL::Joint leftWheelJoint=segmentMapIter->second.segment.getJoint();
		wheelRadius[0]=chassisFrame(2,3)+leftSupportFrame(2,3)+leftWheelJoint.JointOrigin().z();

                segmentMapIter=tree.getSegment("right_wheel");
                KDL::Joint rightWheelJoint=segmentMapIter->second.segment.getJoint();
		wheelRadius[1]=chassisFrame(2,3)+rightSupportFrame(2,3)+rightWheelJoint.JointOrigin().z();
		
		const double K5 = 0.0018533548425194695;
		const double K6 = 0.09946140462774823;
		const double K7 = 21.65458426501294;
		const double K8 = -15.40102896939387;
		Ginv << 0.5*K7, 0.5*K8,
		        0.5*K7, -0.5*K8;
		F << 0.0, K5,
                     K6, 0.0;

		return true;
	}
	
	void CartLinearizingController::starting(const ros::Time& time)
	{
		v.setZero();
	}
	
	void CartLinearizingController::update(const ros::Time& time,
	        const ros::Duration& duration)
	{
		
		for(unsigned int i=0;i < joints_.size();i++)
		{
			nu[i]=joints_[i].getVelocity();
		}
		
		u[0]=(nu[0]*wheelRadius[0]+nu[1]*wheelRadius[1])/2.0;
		u[1]=(nu[0]*wheelRadius[0]-nu[1]*wheelRadius[1])/wheelBase;
		
		// Compute linearization
	        Eigen::Vector2d uf;
	        uf << u[0]*u[1], u[1]*u[1];

		torque=Ginv*(v+F*uf);
		
		// Apply torques
		for(unsigned int i=0;i < joints_.size();i++) joints_[i].setCommand(torque(i));
	}
	
	void CartLinearizingController::commandCB(const std_msgs::Float64MultiArray::ConstPtr &command)
	{
		for(unsigned int i=0;i < command->data.size();i++) v[i]=command->data[i];
	}
}
PLUGINLIB_EXPORT_CLASS(twil_controllers::CartLinearizingController,controller_interface::ControllerBase)
